@@ -30,6 +30,20 @@ class DNSError(Exception):
 
 # DNS codes
 
+def unknown_qtype(name,key,forward):
+    if forward:
+        try:
+            return "TYPE%d" % (key,)
+        except:
+            raise DNSError("%s: Invalid forward lookup: [%s]" % (name,key))
+    else:
+        if key.startswith("TYPE"):
+            try:
+                return int(key[4:])
+            except:
+                pass
+        raise DNSError("%s: Invalid reverse lookup: [%s]" % (name,key))
+
 QTYPE =  Bimap('QTYPE',
         {1:'A', 2:'NS', 5:'CNAME', 6:'SOA', 10:'NULL', 12:'PTR', 13:'HINFO',
                     15:'MX', 16:'TXT', 17:'RP', 18:'AFSDB', 24:'SIG', 25:'KEY',
@@ -41,7 +55,7 @@ QTYPE =  Bimap('QTYPE',
                     61:'OPENPGPKEY', 62:'CSYNC', 63:'ZONEMD', 64:'SVCB',
                     65:'HTTPS', 99:'SPF', 108:'EUI48', 109:'EUI64', 249:'TKEY',
                     250:'TSIG', 251:'IXFR', 252:'AXFR', 255:'ANY', 256:'URI',
-                    257:'CAA', 32768:'TA', 32769:'DLV'}, DNSError)
+                    257:'CAA', 32768:'TA', 32769:'DLV'}, unknown_qtype)
 
 CLASS =  Bimap('CLASS',
                 {1:'IN', 2:'CS', 3:'CH', 4:'Hesiod', 254:'None', 255:'*'},
@@ -709,7 +723,7 @@ class DNSQuestion(object):
 
     def toZone(self):
        return ';%-30s %-7s %s' % (self.qname,CLASS.get(self.qclass),
-                                             QTYPE.get(self.qtype))
+                                             QTYPE[self.qtype])
 
     def __repr__(self):
         return "<DNS Question: '%s' qtype=%s qclass=%s>" % (
@@ -899,7 +913,7 @@ class RR(object):
         else:
             return '%-23s %-7s %-7s %-7s %s' % (self.rname,self.ttl,
                                                 CLASS.get(self.rclass),
-                                                QTYPE.get(self.rtype),
+                                                QTYPE[self.rtype],
                                                 self.rdata.toZone())
 
     def __str__(self):
@@ -1031,8 +1045,10 @@ class RD(object):
         """
             Default 'repr' format should be equivalent to RD zone format
         """
-        # For unknown rdata just default to hex
-        return binascii.hexlify(self.data).decode()
+        if len(self.data) > 0:
+            return "\\# %d %s" % (len(self.data), binascii.hexlify(self.data).decode().upper())
+        else:
+            return "\\# 0"
 
     def toZone(self):
         return repr(self)
@@ -1512,6 +1528,50 @@ class NAPTR(RD):
 
     attrs = ('order','preference','flags','service','regexp','replacement')
 
+class DS(RD):
+    """
+        DS (delegation signer) record as specified in RFC 4034 Section 5.
+        https://www.rfc-editor.org/rfc/rfc4034#section-5
+    """
+
+    key_tag = H('key_tag')
+    algorithm = B('algorithm')
+    digest_type = B('digest_type')
+
+    @classmethod
+    def parse(cls,buffer,length):
+        try:
+            (key_tag,algorithm,digest_type) = buffer.unpack("!HBB")
+            digest = buffer.get(length - 4)
+            return cls(key_tag,algorithm,digest_type,digest)
+        except (BufferError,BimapError) as e:
+            raise DNSError("Error unpacking DS [offset=%d]: %s" %
+                                        (buffer.offset,e))
+
+    @classmethod
+    def fromZone(cls,rd,origin=None):
+        return cls(int(rd[0]),int(rd[1]),int(rd[2]),
+                   binascii.unhexlify("".join(rd[3:]).encode('ascii')))
+
+    def __init__(self,key_tag,algorithm,digest_type,digest):
+        self.key_tag = key_tag
+        self.algorithm = algorithm
+        self.digest_type = digest_type
+        self.digest = _force_bytes(digest)
+
+    def pack(self,buffer):
+        buffer.pack("!HBB",self.key_tag,self.algorithm,self.digest_type)
+        buffer.append(self.digest)
+
+    def __repr__(self):
+        return "%d %d %d %s" % (
+                        self.key_tag,
+                        self.algorithm,
+                        self.digest_type,
+                        binascii.hexlify(self.digest).decode().upper())
+
+    attrs = ('key_tag','algorithm','digest_type','digest')
+
 class DNSKEY(RD):
 
     flags = H('flags')
@@ -1810,6 +1870,21 @@ class HTTPS(RD):
         >>> obj.pack(b)
         >>> b.data == pcap
         True
+
+        # Issue 43: HTTPS reads after RD end 
+        >>> msg = binascii.unhexlify("93088410000100020000000107646973636f726403636f6d0000410001c00c004100010000012c002b0001000001000c0268330568332d323902683200040014a29f80e9a29f87e8a29f88e8a29f89e8a29f8ae8c00c002e00010000012c005f00410d020000012c632834e5632575c586c907646973636f726403636f6d0044d488ce4a5b9085289c671f0296b2b06cffaca28880c57643befd43d6de433d84ae078b282fc2cdd744f3bea2f201042a7a0d6f3e17ebd887b082bbe30dfda100002904d0000080000000")
+        >>> print(DNSRecord.parse(msg))
+        ;; ->>HEADER<<- opcode: QUERY, status: NOERROR, id: 37640
+        ;; flags: qr aa cd; QUERY: 1, ANSWER: 2, AUTHORITY: 0, ADDITIONAL: 1
+        ;; QUESTION SECTION:
+        ;discord.com.                   IN      HTTPS
+        ;; ANSWER SECTION:
+        discord.com.            300     IN      HTTPS   1 . alpn=h3,h3-29,h2 ipv4hint=162.159.128.233,162.159.135.232,162.159.136.232,162.159.137.232,162.159.138.232
+        discord.com.            300     IN      RRSIG   HTTPS 13 2 300 20220919092245 20220917072245 34505 discord.com. RNSIzkpbkIUonGcfApaysGz/rKKIgMV2Q779Q9beQz2ErgeLKC/CzddE876i8gEEKnoNbz4X69iHsIK74w39oQ==
+        ;; ADDITIONAL SECTION:
+        ;; OPT PSEUDOSECTION
+        ; EDNS: version: 0, flags: do; udp: 1232
+
     """
 
     attrs = ('priority', 'target', 'params')
@@ -1822,6 +1897,7 @@ class HTTPS(RD):
     @classmethod
     def parse(cls,buffer,length):
         try:
+            end = buffer.offset + length
             priority, = buffer.unpack("!H")
             target = []
             while True:
@@ -1831,7 +1907,7 @@ class HTTPS(RD):
                 seg = bytearray(buffer.get(n))
                 target.append(seg)
             params = []
-            while buffer.remaining() > 0:
+            while buffer.offset < end:
                 k, = buffer.unpack("!H")
                 n, = buffer.unpack("!H")
                 v = bytearray(buffer.get(n))
@@ -2087,13 +2163,314 @@ class HTTPS(RD):
         targ = ".".join([self.zf_tostr(t) for t in self.target]) + "."
         return " ".join([pri, targ] + [self.zf_format_param(k, v) for k,v in self.params])
 
+class SSHFP(RD):
+    """
+        SSHFP record as specified in RFC 4255
+        https://www.rfc-editor.org/rfc/rfc4255.html
+    """
+
+    algorithm = B('algorithm')
+    fp_type = B('fp_type')
+
+    @classmethod
+    def parse(cls,buffer,length):
+        try:
+            (algorithm,fp_type) = buffer.unpack("!BB")
+            fingerprint = buffer.get(length - 2)
+            return cls(algorithm,fp_type,fingerprint)
+        except (BufferError,BimapError) as e:
+            raise DNSError("Error unpacking DS [offset=%d]: %s" %
+                                        (buffer.offset,e))
+
+    @classmethod
+    def fromZone(cls,rd,origin=None):
+        return cls(int(rd[0]),int(rd[1]),
+                   binascii.unhexlify("".join(rd[2:]).encode('ascii')))
+
+    def __init__(self,algorithm,fp_type,fingerprint):
+        self.algorithm = algorithm
+        self.fp_type = fp_type
+        self.fingerprint = _force_bytes(fingerprint)
+
+    def pack(self,buffer):
+        buffer.pack("!BB",self.algorithm,self.fp_type)
+        buffer.append(self.fingerprint)
+
+    def __repr__(self):
+        return "%d %d %s" % (
+                        self.algorithm,
+                        self.fp_type,
+                        binascii.hexlify(self.fingerprint).decode().upper())
+
+    attrs = ('algorithm','fp_type','fingerprint')
+
+class TLSA(RD):
+    """
+        TLSA record as specified in RFC 6698
+        https://www.rfc-editor.org/rfc/rfc6698
+    """
+
+    cert_usage = B('cert_usage')
+    selector = B('selector')
+    matching_type = B('matching_type')
+
+    @classmethod
+    def parse(cls,buffer,length):
+        try:
+            (cert_usage,selector,matching_type) = buffer.unpack("!BBB")
+            cert_data = buffer.get(length - 3)
+            return cls(cert_usage,selector,matching_type,cert_data)
+        except (BufferError,BimapError) as e:
+            raise DNSError("Error unpacking DS [offset=%d]: %s" %
+                                        (buffer.offset,e))
+
+    @classmethod
+    def fromZone(cls,rd,origin=None):
+        return cls(int(rd[0]),int(rd[1]),int(rd[2]),
+                   binascii.unhexlify("".join(rd[3:]).encode('ascii')))
+
+    def __init__(self,cert_usage,selector,matching_type,cert_data):
+        self.cert_usage = cert_usage
+        self.selector = selector
+        self.matching_type = matching_type
+        self.cert_data = _force_bytes(cert_data)
+
+    def pack(self,buffer):
+        buffer.pack("!BBB",self.cert_usage,self.selector,self.matching_type)
+        buffer.append(self.cert_data)
+
+    def __repr__(self):
+        return "%d %d %d %s" % (
+                        self.cert_usage,
+                        self.selector,
+                        self.matching_type,
+                        binascii.hexlify(self.cert_data).decode().upper())
+
+    attrs = ('cert_usage','selector','matching_type','cert_data')
+
+class LOC(RD):
+    """
+        LOC record as specified in RFC 1876
+
+        >>> LOC(37.236693, -115.804069, 1381.0)
+        37 14 12.094 N 115 48 14.649 W 1381.00m
+        >>> LOC(37.236693, -115.804069, 1381.0, 3000.0, 1.0, 1.0)
+        37 14 12.094 N 115 48 14.649 W 1381.00m 3000.00m 1.00m 1.00m
+        >>> a = DNSRecord(DNSHeader(id=1456))
+        >>> a.add_answer(*RR.fromZone('area51.local. 60 IN LOC 37 14 12.094 N 115 48 14.649 W 1381.00m'))
+        >>> a.add_answer(*RR.fromZone('area51.local. 60 IN LOC 37 N 115 48 W 1381.00m'))
+        >>> a.add_answer(*RR.fromZone('area51.local. 60 IN LOC 37 14 12.094 N 115 48 14.649 W 1381.00m 1m 10000m 10m'))
+        >>> print(a)
+        ;; ->>HEADER<<- opcode: QUERY, status: NOERROR, id: 1456
+        ;; flags: rd; QUERY: 0, ANSWER: 3, AUTHORITY: 0, ADDITIONAL: 0
+        ;; ANSWER SECTION:
+        area51.local.           60      IN      LOC     37 14 12.094 N 115 48 14.649 W 1381.00m
+        area51.local.           60      IN      LOC     37 N 115 48 W 1381.00m
+        area51.local.           60      IN      LOC     37 14 12.094 N 115 48 14.649 W 1381.00m
+    """
+
+    @classmethod
+    def parse(cls, buffer, length):
+        try:
+            (_, siz, hp, vp) = buffer.unpack('!BBBB')
+            (lat, lon, alt) = buffer.unpack('!III')
+            self = cls.__new__(cls)
+            self._lat = lat
+            self._lon = lon
+            self._alt = alt
+            self._siz = siz
+            self._hp = hp
+            self._vp = vp
+            return self
+        except (BufferError, BimapError) as e:
+            raise DNSError("Error unpacking LOC [offset=%d]: %s" %
+                                        (buffer.offset,e))
+
+    @classmethod
+    def fromZone(cls, rd, origin=None):
+        args = []
+        # We still support Python 2.7 so use nonlocal workaround 
+        class context:
+            idx = 0
+        tofloat = lambda x: float(x[:-1])  # get float from "100.0m"
+        def todecimal(chars):
+            decimal = 0.0
+            multiplier = 1
+            for c in chars:
+                if c in rd:
+                    nxt = rd.index(c)
+                    if c in ('S', 'W'):
+                        multiplier = -1
+                    break
+            else:
+                raise DNSError('Missing cardinality [{chars}]'.format(chars=chars))
+            for n, d in zip(rd[context.idx:nxt], (1, 60, 3600)):
+                decimal += float(n) / d
+            context.idx = nxt + 1
+            return decimal * multiplier
+
+        args.append(todecimal('NS'))
+        args.append(todecimal('EW'))
+
+        try:
+            while True:
+                args.append(tofloat(rd[context.idx]))
+                context.idx += 1
+        except IndexError:
+            return cls(*args)
+
+    def __init__(self, lat, lon, alt, siz=1.0, hp=10000.0, vp=10.0):
+        self._lat = int(lat * 3600000 + pow(2, 31))
+        self._lon = int(lon * 3600000 + pow(2, 31))
+        self._alt = int((alt + 100000) * 100)
+        self._siz = LOC.__tosiz(siz)
+        self._hp = LOC.__tosiz(hp)
+        self._vp = LOC.__tosiz(vp)
+
+    def pack(self, buffer):
+        buffer.pack("!BBBB", 0, self._siz, self._hp, self._vp)
+        buffer.pack("!III", self._lat, self._lon, self._alt)
+
+    @property
+    def siz(self):
+        return self.__reprsiz(self._siz)
+
+    @property
+    def hp(self):
+        return self.__reprsiz(self._hp)
+
+    @property
+    def vp(self):
+        return self.__reprsiz(self._vp)
+
+    @property
+    def lat(self):
+        c = 'N' if self._lat > pow(2, 31) else 'S'
+        return self._reprcoord(self._lat, c)
+
+    @property
+    def lon(self):
+        c = 'E' if self._lon > pow(2, 31) else 'W'
+        return self._reprcoord(self._lon, c)
+
+    @property
+    def alt(self):
+        return self._alt / 100 - 100000
+
+    @staticmethod
+    def _reprcoord(value, c):
+        base = abs(pow(2, 31) - value)
+        d = base // 3600000
+        m = base % 3600000 // 60000
+        s = base % 3600000 % 60000 / 1000.0
+
+        if int(s) == 0:
+            if m == 0:
+                return '{d} {c}'.format(c=c,d=d)
+            else:
+                return '{d} {m} {c}'.format(d=d,m=m,c=c)
+        return '{d} {m} {s:.3f} {c}'.format(d=d,m=m,s=s,c=c)
+
+    def __repr__(self):
+        DEFAULT_SIZ = 0x12  # 1m
+        DEFAULT_HP = 0x16   # 10,000m
+        DEFAULT_VP = 0x13   # 10m
+
+        result = '{self.lat} {self.lon} {self.alt:.2f}m'.format(self=self)
+
+        if self._vp != DEFAULT_VP:
+            result += ' {self.siz:.2f}m {self.hp:.2f}m {self.vp:.2f}m'.format(self=self)
+        elif self._hp != DEFAULT_HP:
+            result += ' {self.siz:.2f}m {self.hp:.2f}m'.format(self=self)
+        elif self._siz != DEFAULT_SIZ:
+            result += ' {self.siz:.2f}m'.format(self=self)
+
+        return result
+
+    @staticmethod
+    def __tosiz(v):
+        if int(v) == 0:
+            return 0
+        e = 0
+        v *= 100
+        while v >= 10 and e < 9:
+            v /= 10
+            e += 1
+        v = int(round(v))
+        if v >= 10:
+            raise DNSError("Value out of range")
+        return v << 4 | e
+
+    @staticmethod
+    def __reprsiz(v):
+        b = v >> 4
+        e = v & 0x0F
+        if b > 9 or e > 9 or (b == 0 and e > 0):
+            raise DNSError("Value out of range")
+        return b * pow(10, e) / 100
+
+class RP(RD):
+     """
+     RP record as specified in RFC 1183.
+     https://datatracker.ietf.org/doc/html/rfc1183
+     """
+     @classmethod
+     def parse(cls,buffer,length):
+         try:
+             mbox = buffer.decode_name()
+             txt = buffer.decode_name()
+             return cls(mbox, txt)
+         except (BufferError,BimapError) as e:
+             raise DNSError("Error unpacking RP [offset=%d]: %s" %
+                                         (buffer.offset,e))
+
+     @classmethod
+     def fromZone(cls,rd,origin=None):
+         return cls(label(rd[0],origin),label(rd[1],origin))
+
+     def __init__(self,mbox=None, txt=None):
+         self.mbox = mbox
+         self.txt = txt
+
+     def set_mbox(self,mbox):
+         if isinstance(mbox,DNSLabel):
+             self._mbox = mbox
+         else:
+             self._mbox = DNSLabel(mbox)
+
+     def get_mbox(self):
+         return self._mbox
+
+     mbox = property(get_mbox,set_mbox)
+
+     def set_txt(self,txt):
+         if isinstance(txt,DNSLabel):
+             self._txt = txt
+         else:
+             self._txt = DNSLabel(txt)
+
+     def get_txt(self):
+         return self._txt
+
+     txt = property(get_txt,set_txt)
+
+     def pack(self,buffer):
+         buffer.encode_name(self.mbox)
+         buffer.encode_name(self.txt)
+
+     def __repr__(self):
+         return "%s %s" % (self.mbox,self.txt)
+
+     attrs = ('mbox','txt')
+
 # Map from RD type to class (used to pack/unpack records)
 # If you add a new RD class you must add to RDMAP
 
 RDMAP = { 'CNAME':CNAME, 'A':A, 'AAAA':AAAA, 'TXT':TXT, 'MX':MX,
           'PTR':PTR, 'SOA':SOA, 'NS':NS, 'NAPTR': NAPTR, 'SRV':SRV,
           'DNSKEY':DNSKEY, 'RRSIG':RRSIG, 'NSEC':NSEC, 'CAA':CAA,
-          'HTTPS': HTTPS
+          'HTTPS': HTTPS, 'DS':DS, 'SSHFP':SSHFP, 'TLSA':TLSA, 'LOC':LOC,
+          'RP':RP,
         }
 
 ##
